package roboforge.net;

import roboforge.BDebug;
import roboforge.net.services.*;
import roboforge.net.sockets.*;

import java.io.*;
import java.security.*;



public abstract class BAbstractPresentation implements BPresentationLayer
{
  
  protected BTransportLayer transport;
  protected PrivateKey      key;


  // Constructor...
  // -------------------------------------------------------------

  public BAbstractPresentation(BNetworkStack stack, String keyFile) throws BNetException
  {
    this.key       = getKey(keyFile);
    this.transport = stack.getTransportLayer();
  }


  public BAbstractPresentation(BNetworkStack stack) throws BNetException
    // Alternate constructor used by BClientPresentation as no
    // need to set private key for this class.
  {
    this.transport = stack.getTransportLayer();
  }


  // Message send method...
  // -------------------------------------------------------------

  public void send(BTransmission t)
    throws BNetException
  {
    // Set transmission socket.
    t.setSocket(transport.getSocket(t));

    // Send data.
    sendTransmission(t);
  }


  // Message receive method...
  // -------------------------------------------------------------

  public void receive(BTransmission t)
    throws BNetException
  {
    // Set transmission socket.
    t.setSocket(transport.getSocket(t));

    // Receive data.
    receiveTransmission(t);
  }


  // Message Request method...
  // -----------------------------------------------------------

  public void request(BTransmission t)
    throws BNetException
  {
    // Set transmission socket.
    t.setSocket(transport.getSocket(t));

    // Send data.
    sendTransmission(t);

    // Receive data.
    receiveTransmission(t);
  }


  // Underlying socket transport methods...
  // -------------------------------------------------------------

  protected void sendTransmission(BTransmission t)
    throws BNetException
  {

    try
    {
      // Get the transmission members...
      BSocket       socket  = t.getSocket();
      BTarget       target  = t.getTarget();
      BMessage      message = t.getMessage();
      BDataService  service = t.getSendService();


      // Send security header...
      // ........................................................

      inform(t, "Initiating secure connection (can take up to one minute)...");

      if (service != null)
        service.inform(target);

      sendSecurityHeader(target, socket);


      // DES Encrypt all messaging traffic...
      // ........................................................

      BDESService encryption =
        new BDESService();

      encryption.inform(target);
      socket = encryption.getSocket(socket);


      // Send presentation header...
      // ........................................................

      inform(t, "Sending data to server...");

      BPresentationHeader header =
        new BPresentationHeader();

      // Obtain raw message bytes..
      byte[] data = objectToBytes(message);
      header.setSize(data.length);

      // Place message services in pres. header..
      if (service != null)
        service.configure(header);

      sendPresentationHeader(target, socket, header);


      // Send message...
      // ........................................................

      inform(t, "Sending message to server...");
      inform(t, 0, data.length);

      if (service != null)
        socket = service.getSocket(socket);

      BObjectSocket out =
        new BObjectSocket(socket);

      out.write(data);
      out.close();

      inform(t, "Message sent successfully.");
    }
    catch (BNetException e)
      { throw e; }
    catch (InterruptedIOException e)
      { throw new BNetException("Timed out while receiving message."); }
    catch (IOException e)
    {
      if (t.isStopped())
        throw new BAbortedException("Transmission aborted.");
      else
        throw new BNetException("Could not send message ["+ e+ "]");
    }
    catch (Exception e)
      { throw new BNetException("Internal network error ["+ e+ "]"); }

  }


  protected void receiveTransmission(BTransmission t)
    throws BNetException
  {
    try
    {
      // Get the transmission members...
      BSocket socket  = t.getSocket();
      BTarget target  = t.getTarget();

      // Reset transmission data counts...
      inform(t, 0, 0);

      // Receive security header...
      // ........................................................

      inform(t, "Initiating secure connection...");
      receiveSecurityHeader(target, socket);


      // DES Decrypt all messaging traffic...
      // ........................................................

      BDESService decryption =
        new BDESService();

      decryption.inform(target);
      socket = decryption.getSocket(socket);


      // Receive presentation header...
      // ........................................................

      inform(t, "Receiving data from server...");

      BPresentationHeader header =
        receivePresentationHeader(target, socket);


      // Extract info from header...
      // ........................................................

      BServiceSet service = header.getServices();
      service.inform(target);

      // Receive message...
      // ........................................................

      inform(t, "Receiving message from server...");
      inform(t, header.getSize(), 0);

      socket = service.getSocket(socket);

      BObjectSocket in =
        new BObjectSocket(socket);

      byte[] data = (byte[]) in.read();

      BMessage message =
        (BMessage) bytesToObject(data);

      // Pass received message to the transmission.
      t.setMessage(message);

      inform(t, "Message received successfully.");
    }
    catch (BNetException e)
      { throw e; }
    catch (InterruptedIOException e)
      { throw new BNetException("Timed out while receiving message."); }
    catch (IOException e)
    {
      if (t.isStopped())
        throw new BAbortedException("Transmission aborted.");
      else
        throw new BNetException("Could not send message ["+ e+ "]");
    }
    catch (ClassNotFoundException e)
      { throw new BNetException("Received an unrecognized message."); }
    catch (Exception e)
      { throw new BNetException("Internal network error ["+ e+ "]"); }
  }


  
  // Presentation header methods...
  // -------------------------------------------------------------

  protected void sendPresentationHeader 
  (
    BTarget target, 
    BSocket socket, 
    BPresentationHeader header
  )
    throws IOException
  {
    BObjectSocket out = new BObjectSocket(socket);
    out.write(header);
  }

  
  protected BPresentationHeader receivePresentationHeader
  (
    BTarget target, 
    BSocket socket
  ) 
    throws IOException
  {
    try
    {
      BObjectSocket in = new BObjectSocket(socket);

      BPresentationHeader header =
        (BPresentationHeader) in.read();

      return header;
    }
    catch (ClassNotFoundException e)
      { throw new BNetException("Could not decode presentation header."); }
  }
        
  
        
  // Abstract methods...
  // -------------------------------------------------------------
  
  protected abstract void sendSecurityHeader
    (BTarget target, BSocket socket) throws IOException;
    // Send message security-related information to the given target.
  
  protected abstract void receiveSecurityHeader
    (BTarget target, BSocket socket) throws IOException;
    // Receive message security-related information from the given target.

  protected abstract void inform(BTransmission t, String s);
    // Informs the transmission of message 's'.

  protected abstract void inform(BTransmission t, long in, long out);
    // Informs the transmission of expected I/O quantities.




  // Implementation methods...
  // -------------------------------------------------------------

  protected byte[] objectToBytes(Object o)
  {
    try
    {
      BByteSocket   bytes = new BByteSocket();
      BObjectSocket obj   = new BObjectSocket(bytes);

      obj.write(o);
      obj.close();

      return bytes.getBytes();
    }
    catch (IOException e)
      { return null; }
  }
    
  protected Object bytesToObject(byte[] b)
  {
    try
    {
      BByteSocket   bytes = new BByteSocket(b);
      BObjectSocket obj   = new BObjectSocket(bytes);
          
      return obj.read();
    }
    catch (Exception e)
      { return null; }
  }

  
  protected PrivateKey getKey(String keyFile) throws BNetException
  {
    try
    {
      FileInputStream f =
        new FileInputStream(keyFile);

      ObjectInputStream in =
        new ObjectInputStream(f);

      return (PrivateKey) in.readObject();
    }
    catch (IOException e)
      { throw new BNetException("Could not load Private key: "+ keyFile); }
    catch (ClassNotFoundException e)
      { throw new BNetException("Could not deserialize Private key."); }
  }
  
}
