package roboforge.particles;

import javax.vecmath.*;

import roboforge.particles.emitters.*;
import roboforge.particles.forces.*;
import roboforge.particles.forms.*;
import roboforge.particles.looks.*;
import roboforge.particles.physics.*;


public class BParticle
{

  // Member variables...
  // -------------------------------------------------------------

  public float m;
    // Mass of this particle.

  public Point3f p;
    // 3D position of particle.

  public Point3f oldp;
    // Old 3D position of particle.

  public Vector3f v;
    // 3D velocity of particle.


  // Aging variables...
  // -------------------------------------------------------------

  public float age;
    // Time that this particle lived, in seconds.
    // When age >= lifespan, particle is 'dead'.

  public float lifespan;
    // Life expectancy of this particle, in seconds.

  public float alpha;
    // Age, normalized to the interval [0.0 .. 1.0].



  // Particle aspects...
  // -------------------------------------------------------------

  protected BParticleAspect[] aspects;
    // Aspects of this particle -
    // e.g. physics, look...

  public BParticleForm form;
    // This particle's visual 'form' -
    // i.e. how it appears to users.

    

  // Constructor...
  // -------------------------------------------------------------

  public BParticle()
  {
    this.p        = new Point3f();
    this.oldp     = new Point3f();
    this.v        = new Vector3f();
    this.lifespan = 0f;
    this.aspects  = new BParticleAspect[0];
  }



  // Methods...
  // -------------------------------------------------------------

  public void start()
  {
    this.age      = 0f;
    this.alpha    = 0f;

    // Start particle aspects...
    int n = aspects.length;
    for (int i=0; i<n; i++)
      aspects[i].start(this);
  }

  public void update(float t, float dt)
  {
    // Update particle's aging.
    age  += dt;
    alpha = age / lifespan;

    // Update particle aspects...
    int n = aspects.length;
    for (int i=0; i<n; i++)
      aspects[i].update(this, t, dt);
  }


  public void stop()
  {
    // Stop particle aspects...
    int n = aspects.length;
    for (int i=0; i<n; i++)
      aspects[i].stop(this);
  }



  // Aspect Methods...
  // -------------------------------------------------------------

  public void addAspect(BParticleAspect aspect)
  {
    if (aspect == null)
      return;

    int n = aspects.length;

    // Allocate space for the specified aspect.
    BParticleAspect[] old = aspects;
    aspects = new BParticleAspect[n+1];
    System.arraycopy(old, 0, aspects, 0, n);

    // Store the specified aspect.
    aspects[n] = aspect;
  }

  public int getAspectCount()
  	{ return aspects.length; }

    

  // Accessor methods...
  // -------------------------------------------------------------

  public boolean isAlive()
    { return (alpha <= 1.0f); }

  public boolean isDead()
    { return (alpha >= 1.0f); }

  public void kill()
  	{ age = lifespan; }


  public void setMass(float mass)
    { m = mass; }

  public float getMass()
    { return m; }


  public void setLifeSpan(float span)
  {
    this.lifespan = span;
    this.alpha    = age / span;
  }

  public float getLifeSpan()
    { return this.lifespan; }


  public void setForm(BParticleForm form)
    { this.form = form; }

  public BParticleForm getForm()
    { return this.form; }

}
