package roboforge.particles;

import roboforge.BDebug;
import roboforge.animation.*;
import roboforge.animation.time.*;

import roboforge.util.timing.BClock;

import roboforge.particles.emitters.*;
import roboforge.particles.forces.*;
import roboforge.particles.forms.*;
import roboforge.particles.looks.*;
import roboforge.particles.physics.*;



public abstract class BParticleSystemAbstract
	extends		 BTimeAnimation
	implements BParticleSystem
{


  // System aspects...
  // -------------------------------------------------------------

  protected BSystemAspect[] aspects;
    // A list of particle system aspects -
    // For example, physics, emission, look.


  // Constructor...
  // -------------------------------------------------------------

  public BParticleSystemAbstract()
  {
  	super();
    aspects = new BSystemAspect[0];
  }



  // Animation Methods...
  // -------------------------------------------------------------

  public void start(float t)
    { startAspects(); }

  public boolean animate(float t, float dt)
  {
    try
    {
      // Update system aspects...
      updateAspects(t, dt);

      // Update existing particles...
      updateParticles(t, dt);
    }
    catch (BParticleSystemException e)
    {
      // TODO: Handle error.
      BDebug.print("ParticleSystem error: "+ e);
    }

    return isAlive();
  }

  public void stop(float t)
  {
    stopAspects();
    stopParticles();
  }



  // Aspect Methods...
  // -------------------------------------------------------------

  public void addAspect(BSystemAspect aspect)
  {
    if (aspect == null)
      return;

    int n = aspects.length;

    // Allocate space for the specified aspect.
    BSystemAspect[] old = aspects;
    aspects = new BSystemAspect[n+1];
    System.arraycopy(old, 0, aspects, 0, n);

    // Store the specified aspect.
    aspects[n] = aspect;
  }

  protected void startAspects()
  {
    // Start system aspects...
    int n = aspects.length;

    for (int i=0; i<n; i++)
      aspects[i].start();
  }

  protected void updateAspects(float t, float dt)
  {
    // Update system aspects...
    int n = aspects.length;

    for (int i=0; i<n; i++)
      aspects[i].update(t, dt);
  }

  protected void stopAspects()
  {
    // Stop system aspects...
    int n = aspects.length;

    for (int i=0; i<n; i++)
      aspects[i].stop();
  }


  protected abstract void updateParticles(float t, float dt)
    throws BParticleSystemException;
    // Update all live particles within the system.

  protected abstract void stopParticles();
    // Update all live particles within the system.


  protected BParticle createParticle()
    throws BParticleSystemException
    // Return a single new particle.
  {
    BParticle p = new BParticle();
    configureParticle(p);
    return p;
  }


  protected void configureParticle(BParticle p)
  	throws BParticleSystemException
  {
    // Configure particle aspects...
    int n = aspects.length;
    for (int i=0; i<n; i++)
      p.addAspect(aspects[i].createParticleAspect(p));
  }


  public String toString()
  	{ return getName(); }

  public String getName()
  	{ return "ParticleSystem"; }

}
