package roboforge.ui.modes;

import roboforge.BBot;
import roboforge.BDebug;
import roboforge.BComponent;
import roboforge.BCollisionRecord;
import roboforge.ui.B3DBotCanvas;

import java.awt.event.*;
import java.util.*;



public class BModeManipulator extends BModePick
{

  // Member variables...
  // -------------------------------------------------------------

  protected B3DBotCanvas botcanvas;
    // Bot canvas - used to determine selected component.

  protected CollisionChecker checker;
    // Collision checker object: performs multi-threaded checking.



  // Constructor.
  // -------------------------------------------------------------

  public BModeManipulator(B3DBotCanvas canvas)
  {
    super(canvas);
    this.botcanvas = canvas;
  }


  // Activation methods...
  // -------------------------------------------------------------

  public void onActivate()
  {
    super.onActivate();

    // Create a new collision checker.
    checker = new CollisionChecker();

    // Activate collision checker...
    checker.activate();

    // Check collisions on selected component (if any)...
    checker.setComponent(canvas.getSelectedComponent());
  }

  public void onDeactivate()
  {
    // Deactivate collision checker...
    checker.deactivate();

    // Forget checker.
    checker = null;

    super.onDeactivate();
  }


  // Mouse methods...
  // -------------------------------------------------------------

  public void mouseReleased(MouseEvent e)
  {
    super.mouseReleased(e);

    // Check collisions on selected component (if any)...
    checker.setComponent(canvas.getSelectedComponent());
  }




  // Collision checking class...
  // -------------------------------------------------------------

  protected class CollisionChecker extends Thread
  {
    // This class implements multi-threaded collision checking.
    // The interface consists of three methods:
    // activate(), deactivate() and setComponent().


    protected boolean isActive;
      // Whether this thread is active or not.

    protected BComponent component;
      // Component used to check for collisions.

    protected Vector collisions;
      // Collection of collision records.

    protected Signal test;
      // Object for pending collision test notification.


    public CollisionChecker()
    {
      // Initialize variables...
      isActive    = false;
      component   = null;
      collisions  = new Vector();
      test        = new Signal();
    }

    public synchronized void activate()
    {
      // Set status to active..
      isActive = true;

      // Start multi-threading...
      start();
    }

    public synchronized void deactivate()
    {
      // Set status to inactive.  This should cause the thread to
      // stop running in the near future.
      isActive = false;

      // Set component to null, causing thread to run.
      setComponent(null);
    }


    public synchronized void setComponent(BComponent c)
    {
      // Set the next component...
      component = c;

      // Signal that a new test is waiting.
      test.post();
    }


    public void run()
    {
      while (isActive)
      {
        // Wait for a test posting...
        test.receive();

        // Perform cleanup for previous test...
        clearCollisions();

        // Test the component...
        test(component);
      }

      // Finished: perform a final cleanup...
      clearCollisions();
    }

    protected void test(BComponent c)
    {
      // If no component to test, abort.
      if (c == null) return;

      try
      {
        synchronized (collisions)
        {
          BBot bot = botcanvas.getBot();

          if (isActive)
            // Test bot for self collisions with c.
            bot.testSelfCollisions(collisions, botcanvas.getLocale3d(), c);

          if (isActive)
            // Test for collisions with floor.
            bot.recurseCollisionTestFloor(c, collisions, botcanvas.getLocale3d());

          if (isActive)
            // Show any detected collisions.
            showCollisions();
        }
      }
      catch (Exception e)
        { BDebug.print ("BModeManipulator ERROR: " + e); }
    }



    protected void showCollisions()
    {
      synchronized (collisions)
      {
        int n = collisions.size();
        BCollisionRecord collision;

        for (int i=0; i<n && isActive; i++)
        {
          // Highlight each collision-pair...
          collision = (BCollisionRecord) collisions.get(i);
          collision.setHighlight(true);
        }
      }
    }

    protected void clearCollisions()
    {
      synchronized (collisions)
      {
        BCollisionRecord collision;
        int n = collisions.size();

        for (int i=0; i<n; i++)
        {
          // Remove collision highlights...
          collision = (BCollisionRecord) collisions.get(i);
          collision.setHighlight(false);
        }

        collisions.clear();
      }
    }

  }


  protected class Signal
  {
    protected boolean isSignalled = false;

    public synchronized void post()
    {
      isSignalled = true;
      notify();
    }

    public synchronized void receive()
    {
      while (!isSignalled)
        try
          { wait(); }
        catch (InterruptedException e)
          { }

      isSignalled = false;
    }

  }

}
