package roboforge.util;

import java.util.HashMap;
import java.util.LinkedList;

import java.lang.ref.SoftReference;


public class BSoftCache extends HashMap
{

  // BSoftCache keeps a cache of soft references to objects.
  // That is, they can be nullified at the discretion of the
  // GC algorithm.  You can optionally guarantee that the N
  // most-recently-used objects (MRU) will not get GC'ed.


  // Class variables...
  // -------------------------------------------------------------

  public static final int ALL     = -1;
  public static final int NONE    = 0;
  public static final int DEFAULT = 10;



  // Member variables...
  // -------------------------------------------------------------

  protected LinkedList MRU;
    // Strong references to the most recently used (MRU) entries.

  protected int MRUSize;
    // Number of MRU entries to retain.


  protected boolean retainAll;
    // Retain all entries?

  protected boolean retainNone;
    // Retain all entries?


  protected long hits;
    // Number of cache hits so far.

  protected long misses;
    // Number of cache misses so far.



  // Constructors...
  // -------------------------------------------------------------

  public BSoftCache()
    { this(NONE); }

  public BSoftCache(int mruSize)
  {
    super();

    retainMRU(mruSize);
    this.hits      = 0;
    this.misses    = 0;
  }

  public BSoftCache(int capacity, float loadFactor, int mruSize)
  {
    super(capacity, loadFactor);

    retainMRU(mruSize);
    this.hits      = 0;
    this.misses    = 0;
  }



  // MRU Implementation methods...
  // -------------------------------------------------------------

  protected void updateMRU(Object o)
  {
    if (!retainNone)
    {
      MRU.addFirst(o);

      if (MRU.size() > MRUSize)
        if (!retainAll)
          MRU.removeLast();
    }
  }

  public void retainAll()
    { retainMRU(ALL); }

  public void retainNone()
    { retainMRU(NONE); }

  public void retainMRU(int size)
  {
    if (size < -1)
      throw new IllegalArgumentException("Illegal MRU size requested.");

    this.MRUSize    = size;
    this.retainNone = (size == NONE);
    this.retainAll  = (size == ALL);

    if (retainNone)
    {
      // Count = 0.
      if (MRU != null)
        MRU.clear();

      MRU = null;
    }
    else
    {
      if (MRU == null)
        // Assign the linked list.
        MRU = new LinkedList();

      else if (!retainAll)
      {
        // Clear out any excess cache entries...
        int n = (MRU.size() - MRUSize);

        for (int i=n; --i >= 0; )
          MRU.removeLast();
      }
    }
  }


  // Accessor methods...
  // -------------------------------------------------------------

  public Object put(Object key, Object o)
  {
    if (o == null)
      throw new IllegalArgumentException("Can't put a null object.");

    SoftReference old =
      (SoftReference) super.put(key, new SoftReference(o));

    updateMRU(o);

    if (old != null)
      return old.get();
    else
      return null;
  }

  public Object get(Object name)
  {
    SoftReference r =
      (SoftReference) super.get(name);

    if (r != null)
    {
      Object o = r.get();

      if (o != null)
      {
        hits++;
        updateMRU(o);
      }
      else
        misses++;

      return o;
    }
    else
    {
      misses++;
      return null;
    }
  }

  public double getHitPercentage()
  {
    if (misses == 0)
    {
      if (hits > 0)
        return 100.0;
      else
        return 0.0;
    }
    else
      return (100.0 * hits) / (hits + misses);
  }

  public void clear()
  {
    super.clear();
    if (MRU != null)
      MRU.clear();
  }


}
