package roboforge.util.parser;

import roboforge.BDebug;

import java.io.Reader;
import java.util.Stack;


public abstract class BLALRParser
{

  // Member variables...
  // -------------------------------------------------------------

  protected BLexicalAnalyzer  lex;
  protected ParseStack        stack;
  protected BSymbol           top, input;

  protected boolean           verbose;


  // Parse tables...
  // -------------------------------------------------------------

  protected Rule[]            rules;
  protected ParseAction[][]   actions;
  protected int[][]           go_to;


  // Constructor...
  // -------------------------------------------------------------

  public BLALRParser(BLexicalAnalyzer lex)
  {
    this.lex     = lex;
    this.stack   = new ParseStack();
    this.verbose = false;

    initRulesTable();
    initActionTable();
    initGoToTable();
  }


  // Table initialization method.
  // -------------------------------------------------------------

  protected abstract void initRulesTable();
  protected abstract void initActionTable();
  protected abstract void initGoToTable();


  // Interface methods...
  // -------------------------------------------------------------

  protected void parseInput(Reader reader) throws BParseException
  {
    try
    {
      // Setup lexical analyzer...
      lex.setInput(reader);
      input = lex.nextSymbol();

      // Initialize parser-state...
      ParseAction action;

      // Setup parse-stack...
      stack.clear();
      stack.push(BSymbol.START_SYMBOL);

      // Declare loop variables...
      boolean done = false;

      while (!done)
      {
        top   = stack.peek();
        input = lex.peekSymbol();

        if (verbose)
          BDebug.print("INPUT: "+ input+ ".  STACK: "+ stack);

        try
          { action = actions[top.state][input.id]; }
        catch
          (ArrayIndexOutOfBoundsException ex)
          { throw new BParseException("Malformed action table."); }

        try
          { action.perform(); }
        catch (BAcceptException ex)
          { done = true; }
      }
    }
    catch (BParseException ex)
      { throw ex; }
    catch (Exception ex)
      { throw new BParseException("Internal error:"+ ex); }
  }


  // ParseAction class...
  // -------------------------------------------------------------

  protected abstract class ParseAction
  {
    public abstract void perform()
      throws BParseException;
  }

  protected class Shift extends ParseAction
  {
    protected int state;

    public Shift(int state)
      { this.state = state; }

    public void perform() throws BParseException
    {
      if (verbose) BDebug.print("SHIFT: "+ input);

      input.state = this.state;
      stack.push(input);
      lex.nextSymbol();
    }
  }

  protected class Reduce extends ParseAction
  {
    protected Rule rule;

    public Reduce(Rule rule)
      { this.rule = rule; }

    public void perform() throws BParseException
    {
      if (verbose) BDebug.print("REDUCE: "+ rule);
      BSymbol lhs = rule.apply();
      top = stack.peek();

      try
        { lhs.state = go_to[top.state][rule.go_to]; }
      catch
        ( Exception ex )
        { throw new BParseException("Malformed goto table."); }

      stack.push(lhs);
    }
  }

  protected class Accept extends ParseAction
  {
    public Accept()
      { super(); }

    public void perform() throws BParseException
      { throw new BAcceptException(); }
  }

  protected class Error extends ParseAction
  {
    public Error()
      { super(); }

    public void perform() throws BParseException
      { throw new BParseException("Syntax error"); }
  }


  // Parser-accept exception - signals parse completion.
  // -------------------------------------------------------------

  protected class BAcceptException extends BParseException
  {
    public BAcceptException()
      { super("Input accepted."); }
  }



  // Grammar-rule class...
  // -------------------------------------------------------------

  protected static abstract class Rule
  {
    protected int go_to;
      // Index of LHS in goto table.

    public Rule(int go_to)
      { this.go_to = go_to; }

    public abstract BSymbol apply() throws BParseException;
      // Pop off RHS symbols.
      // Construct LHS symbol
      // return LHS.
  }



  // Parse-stack class...
  // -------------------------------------------------------------

  protected static class ParseStack
  {
    protected Stack stack;

    public ParseStack()
      { stack = new Stack(); }

    public BSymbol pop()
      { return (BSymbol) stack.pop(); }

    public BSymbol peek()
      { return (BSymbol) stack.peek(); }

    public void push(BSymbol s)
      { stack.push(s); }

    public void clear()
      { stack.clear(); }

    public boolean empty()
      { return stack.empty(); }

    public String toString()
    {
      int n = stack.size();
      String s = "";

      for (int i=0; i<n; i++)
      {
        s += stack.get(i);
        if (i < n-1) s += ", ";
      }

      return s;
    }
  }

}