package roboforge;

import roboforge.BGlobal;
import roboforge.BRenderer;

import roboforge.animation.*;
import roboforge.animation.animation.b3d.*;
import roboforge.animation.time.*;

import roboforge.math.*;
import roboforge.particles.*;
import roboforge.particles.systems.*;
import roboforge.scalability.*;
import roboforge.sound.*;
import roboforge.util.graphics.*;

import java.util.*;
import javax.vecmath.*;
import javax.media.j3d.*;



public class BBattleFX
  extends    BManualAnimator
  implements BScalable
{

  protected BRenderer renderer;
    // Renderer for displaying particle fx..

  protected BranchGroup world;
    // Node to attach particle fx to.

  protected BScalabilityMetric scaling;
    // Metric for scaling performance.

  protected float timeScale = 1f;
    // Time scale - ratio of fx time to real-time.



  protected HashMap smokemap = new HashMap();
    // Collection of smoke particle systems...

  protected static final int SMOKE_COUNT = 2;

  protected BBotSmokeSystem[] smokes;
    // Bot-smoke systems.

  protected int nextFreeSmoke = 0;
    // Index of next free bot-smoke system.


  protected HashMap bot_explosions = new HashMap();
    // Collection of bot-explosion systems...

  protected static final int BIG_EXPLOSION_COUNT = 1;
    // Number of explosion systems available...

  protected BBigExplosion[] big_explosions;
    // Big explosions.

  protected static final int EXPLOSION_COUNT = 3;
    // Number of explosion systems available...

  protected BFlameBallSystem[] explosions;
    // Explosion system pool.

  protected static final int SPARK_COUNT = 4;
    // Number of spark systems available...

  protected BCollisionSystem[] sparks;
    // Sparks system pool - used for collisions.

  protected static final int VORTEX_COUNT = 2;
    // Number of vortex systems available...

  protected BVortexSystem[] vortices;
    // Vortex system pool - used for deadlocks.

  protected static final float DEADLOCK_DURATION = 4f;
    // Length of deadlock animation, in seconds.

  protected BFireworksDisplay fireworks;
    // Fireworks, for end of battle.




  public BBattleFX(BArenaCanvas canvas)
  {
    super();

    this.renderer = canvas;
    this.world    = canvas.getPickGroup();
    this.scaling  = canvas.getFPSMetric();

    // Create regular explosions...
    explosions = new BFlameBallSystem[EXPLOSION_COUNT];
    for (int i=0; i<EXPLOSION_COUNT; i++)
      explosions[i] = new BFlameBallSystem(world, scaling);

    if (!BGlobal.isParticleQualityLow())
    {
      // Set explosion scaling...
      if (BGlobal.isParticleQualityHigh())
        for (int i=0; i<EXPLOSION_COUNT; i++)
          explosions[i].setScalabilityRange(1, 1, 2);

      // Create big explosions...
      big_explosions = new BBigExplosion[BIG_EXPLOSION_COUNT];
      for (int i=0; i<BIG_EXPLOSION_COUNT; i++)
      big_explosions[i] = new BBigExplosion(world, scaling);

      // Create sparks...
      sparks = new BCollisionSystem[SPARK_COUNT];
      for (int i=0; i<SPARK_COUNT; i++)
        sparks[i] = new BCollisionSystem(world, scaling);

      // Create vortices...
      vortices = new BVortexSystem[VORTEX_COUNT];
      for (int i=0; i<VORTEX_COUNT; i++)
      	vortices[i] = new BVortexSystem(world, scaling);

      // Create fireworks...
      fireworks = new BFireworksDisplay(world, scaling);

      // Create bot smokes...
      smokes = new BBotSmokeSystem[SMOKE_COUNT];
      for (int i=0; i<SMOKE_COUNT; i++)
      	smokes[i] = new BBotSmokeSystem(world, scaling);

    }
  }


  public synchronized void add(BAnimation a)
  {
    // Overridden to add synchronization.
    super.add(a);
  }

  public synchronized boolean animate()
  {
    // Overridden to add synchronization.
    return super.animate();
  }

  public synchronized void remove(BAnimation a)
  {
    // Overridden to add synchronization.
    super.remove(a);
  }

  public synchronized void clear()
  {
    // Overridden to add synchronization.
    super.clear();
  }



  public void cleanse()
  {
    // TODO: Cleanse each particle system.
    smokemap.clear();
  }



  public void setScalabilityMetric(BScalabilityMetric scaling)
    { this.scaling = scaling; }

  public BScalabilityMetric getScalabilityMetric()
    { return scaling; }


  public synchronized void spawnExplosion(BComponent c, boolean playSoundFX)
  {
    // SCALABILITY: Always play explosion fx..

    // Play explosion sound if requested...
    if (playSoundFX)
    {
      BSound s = c.getMaterial().getDestructSound();
      if (s != null)
        if (!s.isRunning())
          s.play();
    }

    BBigExplosion s = null;

    // If particles are enabled, try to use a big boom.
    if (!BGlobal.isParticleQualityLow())
      s = getFreeBigExplosion();

    if (s != null)
      { s.setComponent(c); add(s); }
    else
    {
      // No free big booms - try a small boom.
      BFlameBallSystem f = getFreeExplosion();

      if (f != null)
        { f.setComponent(c); add(f); }
    }
  }


  public synchronized void spawnCollision(BCollisionRecord c, boolean playSoundFX)
  {

    // If collision has been nullified, return...
    if (c.getNullified()) return;

    // Some collisions don't have an origin...
    if (c.getCollisionPoint() == null) return;

    // If collision has no intensity, return...
    if (c.getCollisionIntensity() <= 0f) return;

    // Valid collision; play sound fx...
    if (playSoundFX)
      c.playSound();


    // Preferences: if user has set 'no particles', don't do sparks.
    if (BGlobal.isParticleQualityLow()) return;

    // SCALABILITY: If we are running too slow, no spark fx..
    if (scaling.getMetric() < 0.25f) return;

    // Let's create some visual FX...
    BCollisionSystem s = getFreeSpark();

    // Spawn system, if available...
    if (s != null)
      { s.configure(c); add(s); }
  }


  public synchronized void spawnSmoke(BBot bot)
  {
    // Preferences: if user has set 'no particles', don't do smoke.
    if (BGlobal.isParticleQualityLow())
      return;

    // If bot is destroyed, don't do smoke.
    if (bot.getPercentageDamageLevel() >= 100f)
      return;

    BBotSmokeSystem smoke =
      (BBotSmokeSystem) smokemap.get(bot);

    if (smoke == null)
    {
      smoke = smokes[nextFreeSmoke++];
      smoke.setBot(bot);
      smoke.setRate(timeScale);

      // Associate smoke with bot.
      smokemap.put(bot, smoke);
    }

    if (!smoke.isAnimating())
      add(smoke);
  }



  public synchronized void spawnBotExplosion(BBot bot)
  {
    // If bot is already exploded, don't make any fx.
    if (bot.getTransparency() > 0f) return;

    BChassis c = bot.getChassis();

    BParticleSystem explosion =
      (BParticleSystem) bot_explosions.get(bot);

    if (explosion == null)
    {
      if (!BGlobal.isParticleQualityLow())
        // If user wants particles, make a really big boom.
        explosion = new BSuperExplosion(world, scaling, c);
      else
        // Otherwise, make a small boom.
        explosion = new BFlameBallSystem(world, scaling, c);

      bot_explosions.put(bot, explosion);
    }

    if (!explosion.isAnimating())
    {
      // Spawn the explosion...
      add(explosion);

      // Fade the bot out...
      spawnBotFadeout(bot, 0.5f);

      // Play bot destruction sound...
      BSound sound = c.getMaterial().getDestructSound();

      if (sound != null)
        if (!sound.isRunning())
          sound.play();
    }
  }

  public synchronized void spawnBotFadeout(BBot bot, float duration)
    { add(new BFadeBotAnimation(bot, duration)); }

  public synchronized void spawnBotDefeat(BBot bot)
    { add(new BDelayedAnimation(new BFadeBotAnimation(bot, 1f), 2f)); }


  public synchronized void spawnDeadlock(BBot a, BBot b)
  {
    // Preferences: if user has set 'no particles',
    // don't do deadlock vortices.
    if (BGlobal.isParticleQualityLow()) return;

    spawnVortex(a, DEADLOCK_DURATION);
    spawnVortex(b, DEADLOCK_DURATION);
  }


  public synchronized void spawnVortex(BBot bot, float duration)
  {
    // Preferences: if user has set 'no particles', no vortex.
    if (BGlobal.isParticleQualityLow()) return;

    BVortexSystem v;

    for (int i=0; i<vortices.length; i++)
    {
      v = vortices[i];

      if (!v.isAnimating())
      {
	    	v.setComponent(bot.getChassis());
        v.setDuration(duration);
  	    add(v);
        break;
      }
    }
  }

  public synchronized void spawnFireworks()
  {
    if (!BGlobal.isParticleQualityLow())
      add(fireworks);
  }




  protected BCollisionSystem getFreeSpark()
  {
    // Find a free spark system...
    for (int i=0; i<SPARK_COUNT; i++)
      if (!sparks[i].isAnimating())
        return sparks[i];

    // No free systems...
    return null;
  }

  protected BFlameBallSystem getFreeExplosion()
  {
    // Locate a free explosion...
    for (int i=0; i<EXPLOSION_COUNT; i++)
      if (!explosions[i].isAnimating())
        return explosions[i];

    // No free systems...
    return null;
  }

  protected BBigExplosion getFreeBigExplosion()
  {
    // Locate a free explosion...
    for (int i=0; i<BIG_EXPLOSION_COUNT; i++)
      if (!big_explosions[i].isAnimating())
        return big_explosions[i];

    // No free systems...
    return null;
  }



  protected class BBigExplosion extends BCompoundSystem
  {
    protected BFlameBallSystem flame;
    protected BDebrisSystem debris;

    public BBigExplosion(BranchGroup world, BScalabilityMetric scaling)
      { this(world, scaling, null); }

    public BBigExplosion
    (
      BranchGroup world,
      BScalabilityMetric scaling,
      BComponent c
    )
    {
      super();
      flame  = new BFlameBallSystem(world, scaling);
      debris = new BDebrisSystem(world, scaling);

      if (BGlobal.isParticleQualityHigh())
        flame.setScalabilityRange(1, 2, 4);

      BBigExplosion.this.add(flame);
      BBigExplosion.this.add(debris);

      if (c != null)
        setComponent(c);
    }

    public void setComponent(BComponent c)
    {
      flame.setComponent(c);
      debris.setComponent(c);
    }

    public void setRate(float rate)
    {
      flame.setRate(rate);
      debris.setRate(rate);
    }
  }


  protected class BSuperExplosion extends  BBigExplosion
  {

    public BSuperExplosion(BranchGroup world, BScalabilityMetric scaling)
      { this(world, scaling, null); }
      
    public BSuperExplosion
    (
      BranchGroup world,
      BScalabilityMetric scaling,
      BComponent c
    )
    {
      super(world, scaling, c);

      if (BGlobal.isParticleQualityHigh())
        debris.setScalabilityRange(1, 20, 50);
      else
        debris.setScalabilityRange(1, 20, 20);

      debris.setIntensity(2f);
    }
  }

  protected class BCollisionSystem extends BCompoundSystem
  {
    protected BSparkSystem spark;
    protected BSmokeSystem smoke;

    public BCollisionSystem(BranchGroup world, BScalabilityMetric scaling)
    {
      super();
      spark = new BSparkSystem(world, scaling);
      smoke = new BSmokeSystem(world, scaling);
      BCollisionSystem.this.add(spark);
      BCollisionSystem.this.add(smoke);
     }

    public void configure(BCollisionRecord c)
    {
      float intensity =
        c.getCollisionIntensity();

      spark.setOrigin(c.getCollisionPoint());
      spark.setIntensity(intensity);
      smoke.setOrigin(c.getCollisionPoint());
      smoke.setIntensity(intensity);
    }

    public void setRate(float rate)
    {
      spark.setRate(rate);
      smoke.setRate(rate);
    }
  }


  protected class BFireworksDisplay extends BAbstractAnimation
  {
    protected Point3f base   = new Point3f(0, 400, 0);
    protected Point3f vary   = new Point3f(500, 100, 500);
    protected Point3f target = new Point3f();

    protected Color4f cbase  = new Color4f(0.5f, 1.0f, 0.9f, 1f);
    protected Color4f cvary  = new Color4f(1.0f, 0.5f, 0.0f, 0f);
    protected Color4f color  = new Color4f();
    protected Color4f cdiff  = new Color4f(0.2f, 0.2f, 0.2f, 0f);

    protected BFireworkSystem[] firework;

    protected int FIREWORK_COUNT = 2;

    protected BSound bang
      = BSound.getSound(BGlobal.SOUNDPATH+ "explosions/fireworks.wav");



    public BFireworksDisplay(BranchGroup world, BScalabilityMetric scaling)
    {
      super();

      // Instantiate particle systems...
      firework = new BFireworkSystem[FIREWORK_COUNT];

      for (int i=0; i<FIREWORK_COUNT; i++)
        firework[i] = new BFireworkSystem(world, scaling);

      // Configure the bang sound...
      bang.setPriority(BSound.PRIORITY_MIN);
    }

    public boolean animate()
    {
      for (int i=0; i<FIREWORK_COUNT; i++)
        if (!firework[i].isAnimating())
        {
          configure(firework[i]);
          add(firework[i]);
          bang.play();
        }

      return true;
    }

    protected void configure(BFireworkSystem system)
    {
      BMath3D.randomize(base, vary, target);
      system.setOrigin(target);
      system.setIntensity(BMath.random(0.5f, 1f));

      BColors.randomHSB(cbase, cvary, color);
      system.setColors(color, cdiff);
    }

    public void setRate(float rate)
    {
      for (int i=0; i<FIREWORK_COUNT; i++)
        firework[i].setRate(rate);
    }
    
  }




  public void setTimeScale(float scale)
  {
    this.timeScale = scale;

    // Set explosion scales...
    for (int i=0; i<EXPLOSION_COUNT; i++)
      explosions[i].setRate(scale);

    if (!BGlobal.isParticleQualityLow())
    {
      // Set big explosions...
      for (int i=0; i<BIG_EXPLOSION_COUNT; i++)
        big_explosions[i].setRate(scale);

      // Set sparks...
      for (int i=0; i<SPARK_COUNT; i++)
        sparks[i].setRate(scale);

      // Set vortices...
      for (int i=0; i<VORTEX_COUNT; i++)
      	vortices[i].setRate(scale);

      // Set fireworks...
      fireworks.setRate(scale);

      // Set smokes...
      for (int i=0; i<SMOKE_COUNT; i++)
      	smokes[i].setRate(scale);

    }
  }


}
