package roboforge.animation.time;

import roboforge.animation.*;
import roboforge.util.timing.*;



public abstract class BTimeAnimation extends BAbstractAnimation
{

  // Constants...
  // -------------------------------------------------------------

  public static final float DEFAULT_RATE = 1.0f;


  // Members...
  // -------------------------------------------------------------

  protected float rate;
    // How many seconds of animation time correspond
    // to one second of real time.
    // rate = 2.0 => animation runs @ twice real-time.
    // rate = 0.5 => animation runs @ half real-time.

  protected float time;
    // Time elapsed (seconds) in animation, normalised to
    // time = 0 when start() is called.

  private float start, last;
    // Working members for tracking animation time.




  // Constructors...
  // -------------------------------------------------------------

  public BTimeAnimation()
    { this(DEFAULT_RATE); }

  public BTimeAnimation(float rate)
  {
    setRate(rate);
  }


  // Accessor methods...
  // -------------------------------------------------------------

  public void setRate(float rate)
    { this.rate = rate; }

  public float getRate()
    { return rate; }

  public float getTime()
    { return time; }


  // BAnimation implementation...
  // -------------------------------------------------------------


  public synchronized final void start()
  {
    // Early out if start() is called during animation.
    if (isAnimating()) return;

    // Call super.start() to update animation status.
    super.start();

    time  = 0f;
    start = last = (float) BClock.time();

    // Start time-based animation...
    start(time);
  }


  public final boolean animate()
  {
    // Early out if animate() is called while not animating.
    if (!isAnimating()) return false;

    // Calculate current time step.
    float t  = (float) BClock.time();
    float dt = (t - last) * rate;

    // Perform time-based animation...
    boolean result = animate(time, dt);

    // Update timing information...
    time += dt;
    last  = t;

    // Return the result of animation.
    return result;
  }


  public synchronized final void stop()
  {
    // Early out if stop() is called while not animating.
    if (!isAnimating()) return;

    // Call super.stop() to update animation status.
    super.stop();

    // Stop time-based animation...
    stop(time);
  }


  // Abstract time-animation methods...
  // -------------------------------------------------------------

  public abstract void start(float t);
    // Start animation.  t is the time, in secs.
    // t should always be 0.

  public abstract boolean animate(float t, float dt);
    // Returns the result of animating at time t,
    // for dt seconds.

  public abstract void stop(float t);
    // Stop animation.  t is the time, in secs.


}