package roboforge.net;

import java.io.*;

import roboforge.BDebug;
import roboforge.BProgressMonitorable;

import roboforge.net.services.*;
import roboforge.net.sockets.*;




public class BTransmission implements BProgressMonitorable
{

  // Constants...
  // -------------------------------------------------------------

  public static final long NO_TIMEOUT = -1;
    // Specifies that no timeout should occur.

  public static final long DEFAULT_TIMEOUT = 30000;
    // Default timeout value for transmissions.

  public static final int DEFAULT_ATTEMPTS = 3;
  	// Default number of transmission attempts.

  public static final int DEFAULT_ATTEMPT_INTERVAL = 1000;
  	// Default interval between transmission attempts.




  // Member variables...
  // -------------------------------------------------------------

  protected BMessage message;
    // The transmission data.

  protected BTarget target;
    // The target with which to communicate.

  protected BDataService service;
    // Data services to use for sending.



  // Network implementation...
  // -------------------------------------------------------------

  protected BSocket socket;
    // The transmission socket.

  protected BMonitorService monitor;
    // Service used to monitor data transmision.
    // Provides timeout/abort/progress services.

  protected boolean isStopped;
    // Whether this transmission has been stopped or not..

  protected long timeout;
    // Timeout period, in ms.

  protected int attempts;
  	// Number of times client will retry a message
    // before returning a net error.

  protected long attemptInterval;
  	// Interval (in ms) between successive attempts.



  // Constructors...
  // -------------------------------------------------------------

  public BTransmission(BMessage m, long t)
  {
    this();
    setMessage(m);
    setTimeout(t);
  }

  public BTransmission(BMessage m)
  {
    this();
    setMessage(m);
  }

  public BTransmission()
  {
    this.timeout         = DEFAULT_TIMEOUT;
    this.attempts 			 = DEFAULT_ATTEMPTS;
    this.attemptInterval = DEFAULT_ATTEMPT_INTERVAL;
    this.isStopped       = false;
  }


  // Accessor methods...
  // -------------------------------------------------------------

  public void setMessage(BMessage m)
    { this.message = m; }

  public BMessage getMessage()
    { return message; }

  public void setTarget(BTarget t)
    { this.target = t; }

  public BTarget getTarget()
    { return target; }

  public void setSendService(BDataService s)
    { this.service = s; }

  public BDataService getSendService()
    { return service; }


  // Network service accessors...
  // -------------------------------------------------------------

  public void setTimeout(long ms)
    { this.timeout = ms; }

  public long getTimeout()
    { return timeout; }

  public void setAttempts(int attempts)
    { this.attempts = attempts; }

  public int getAttempts()
    { return attempts; }

  public void setAttemptInterval(int ms)
    { this.attemptInterval = ms; }

  public long getAttemptInterval()
    { return attemptInterval; }



  public void setSocket(BSocket s) throws BNetException
  {
    try
    {
      // Instantiate a new monitor...
      monitor = new BMonitorService();

      // Wrap the supplied socket.
      socket = monitor.getSocket(s);
    }
    catch (IOException e)
      { throw new BNetException(e.getMessage()); }
  }

  public BSocket getSocket() throws BNetException
  {
    if (isStopped)
      throw new BNetException("Transmission aborted.");

    return socket;
  }



  // Transmission abort methods...
  // -------------------------------------------------------------

  public void stop()
    // Abort this transmission ASAP.
  {
    isStopped = true;

    if (monitor != null)
    {
      monitor.setStopped(true);
      monitor = null;
    }
  }

  public boolean isStopped()
    { return isStopped; }



  // BProgressMonitorable implementation...
  // -------------------------------------------------------------

  public String getProgressMessage()
  {
    if (monitor != null)
      return monitor.getProgressMessage();
    else
      return "";
  }

  public int getProgressPercent()
  {
    if (monitor != null)
      return monitor.getProgressPercent();
    else
      return 0;
  }

  public void inform(BServiceInfo info)
  {
    // Pass on service info to the monitor.
    if (monitor != null)
      monitor.inform(info);
  }

  public String toString()
    { return getMessage().toString(); }

}
