package roboforge.net.client;

import roboforge.BGlobal;
import roboforge.BDebug;

import roboforge.net.*;
import roboforge.net.data.*;
import roboforge.net.services.*;
import roboforge.net.sockets.*;

import iaik.security.provider.IAIK;

import java.io.*;
import java.security.*;
import java.util.*;
import java.text.DateFormat;



public class BClientStack extends BNetworkStack
{



  // Member variables...
  // -------------------------------------------------------------

  protected String user;
    // Login name for the user that opens this session.

  protected String pass;
    // Password for the user that opens this session.

  protected boolean isLoggedIn;
    // Is the user logged in for this session?

  protected Long accountID;
    // Current user's account ID.

  protected String emailAddress;


  // Security information...
  // -------------------------------------------------------------

  protected int minuteOffSet;
    // Offset of client time from server time, in mins.

  protected TimeZone serverTimeZone;
    // The server's time-zone.


  // Internal variables...
  // -------------------------------------------------------------

  protected BServerTarget server;
    // Reference to server.


  // Constructors...
  // -------------------------------------------------------------

  public BClientStack() throws BNetException
  {
    super();

    IAIK.addAsProvider(false);

    this.isLoggedIn 	 	 = false;
    this.server     	 	 = new BServerTarget();

    // At present, client receives no public-key encrypted msgs.
    // this.publicKey  = getPublicKey(BGlobal.PUBLIC_KEY_CLIENT);

  }


  // Layer factory methods...
  // -------------------------------------------------------------

  protected BPresentationLayer createPresentationLayer()
    throws BNetException
    { return new BClientPresentation(this); }

  protected BTransportLayer createTransportLayer()
    throws BNetException
    { return new BClientTransport(this); }



  // Messaging methods...
  // -------------------------------------------------------------

  public synchronized void login(String user, String pass)
    throws BNetException
  {
    if (isLoggedIn)
      return;

    this.user = user;
    this.pass = pass;

    try
    {
      // Create login request transmission...
      BRequest r      = new BRequest(this, "LOG", null);
      BTransmission t = new BTransmission(r);

      // Get server's response to login request...
      request(t);
      BMessage m = t.getMessage();

      // Determine what to do, based on response.
      if (m == null)
        throw new BNetException("No response from server.");

      else if (!m.isAcknowledgement())
        throw new BNetException(m.getData());

      else if (!(m.getData() instanceof BLoginData))
        throw new BNetException("Server did not respond correctly to login request.");

      else
      {
        // Login acknowledged by server...
        isLoggedIn  = true;

        // Process the login response...
        processLoginData((BLoginData) m.getData());
      }
    }
    catch (BNetException e)
    {
      logout();
      throw e;
    }
    catch (Exception e)
    {
      logout();
      throw new BNetException("Internal network error: "+ e);
    }
  }

  public synchronized void logout()
    throws BNetException
  {
    // if (!isLoggedIn)
    //   return;

    this.isLoggedIn = false;
    this.user       = null;
    this.pass       = null;
  }

  protected void processLoginData(BLoginData data)
  {
    // Determine relative server time...
    long currentTime	= System.currentTimeMillis();
    long serverTime 	= data.serverTime.getTime().getTime();
    long deltaTime 		= currentTime - serverTime;
    double sec 				= (double) deltaTime / 1000.0;

    // Store time information.
    serverTimeZone 		= data.serverTime.getTimeZone();
    minuteOffSet 			= (int) java.lang.Math.round(sec / 60.0);

    // Remember account ID.
    accountID = new Long(data.accountID);

    // Remember email address.
    emailAddress = data.emailAddress;
  }




  public BMessage request(BTransmission t)
    throws BNetException
  {
    // Set transmission target to the server.
    t.setTarget(server);

    // WARNING: SJS [12/06/2001] - Following line of code
    // could cause problems when multiple simultaneous
    // transmissions are processed.  The problems would
    // occur because the server object is shared between
    // all transmissions.  Bad Hugh!

    // Adds a query to the URL being the request type.
    // This is for server side redirection of requests.

    server.resetURLToType(t.getMessage().getType());


    // SJS [09/04/2001] - Added attempts to requests.
    Exception exception = null;

    // SJS [13/06/2001] - Get attempt info from transmission.
    int  attempts = t.getAttempts();
    long interval = t.getAttemptInterval();

    for (int i=0; i<attempts; i++)
    	try
      {
      	// Process client's request.
      	presentation.request(t);

        // Return the response.
        return t.getMessage();
      }

      catch (BAbortedException e)
        // Don't retry if transmission was aborted...
        { throw e; }

      catch (Exception e)
      {
      	// Remember the exception.
      	exception = e;

        // Sleep for a while...
        sleep(interval);
      }

    // If we failed all attempts, throw last exception.
    throw new BNetException
    	("Request failed after "+ attempts+ " attempts. ("+ exception+ ")");
  }



  // Accessor methods...
  // -----------------------------------------------------------

  public synchronized boolean isLoggedIn()
    { return isLoggedIn; }


  public String getUserName()
    { return user; }

  public String getPassword()
    { return pass; }

  public void setPassword(String pass)
    { this.pass = pass; }

  public void setServerTarget(BServerTarget server)
    { this.server = server;}

  public BServerTarget getServerTarget ()
    { return server;}

  // public PublicKey getPublicKey()
  //   { return publicKey; }


  public Long getAccountID()
  {
    if (isLoggedIn)
      return accountID;
    else
      return null;
  }

  public int getServerMinuteOffset()
	  { return minuteOffSet; }

  public TimeZone getServerTimeZone()
    { return serverTimeZone; }



  // Implementation methods...
  // -----------------------------------------------------------

  /*
  protected PublicKey getPublicKey(String keyFile)
  {
    try
    {
      FileInputStream f =
        new FileInputStream(keyFile);

      ObjectInputStream in =
        new ObjectInputStream(f);

      return (PublicKey) in.readObject();
    }
    catch (IOException e)
      { } // ??
    catch (ClassNotFoundException e)
      { } // ??

    return null;
  }
  */


  protected void sleep(long ms)
  {
    // Put the calling thread to sleep for a while.
  	try
	    { Thread.sleep(ms); }
    catch (InterruptedException e)
    	{ }
  }

  public boolean isStubEmail()
  {
    BDebug.print("BClientStack isStubEmail email " +   emailAddress);

    if ( emailAddress == null)
      return false;


    return emailAddress.toLowerCase().indexOf(BGlobal.STUB_EMAIL_SUFFIX) != -1;
  }

}