package roboforge.particles.forms;

import roboforge.BGlobal;
import roboforge.particles.*;

import javax.vecmath.*;
import javax.media.j3d.*;


public abstract class BParticleFormShape implements BParticleForm
{

  protected boolean isAdditive;
    // Whether this form uses additive blending.

  protected boolean isAttached;
  	// Whether this form has been attached to a
    // parent BranchGroup or not..


  protected BranchGroup parent;
  protected BranchGroup root;
  protected TransformGroup tg;
  protected Transform3D transform = new Transform3D();

  protected Shape3D shape;

  protected Appearance appearance;
  protected Material material;
  protected TransparencyAttributes ta;
  protected RenderingAttributes ra;
  protected ColoringAttributes coloring;


  public BParticleFormShape(BranchGroup parent)
    { this(parent, true); }

  public BParticleFormShape(BranchGroup parent, boolean isAdditive)
  {
    this.parent = parent;

    tg = new TransformGroup();
    tg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);

    shape = createShape3D();
    tg.addChild(shape);

    createAppearance();
    shape.setAppearance(appearance);

    // Must be done after appearance has been created...
    this.setAdditive(isAdditive);

    root = new BranchGroup();
    root.addChild(tg);

    // Compile the scene subgraph...
    root.compile();

    this.isAttached = false;
  }


  protected abstract Shape3D createShape3D();
    // Returns a Geometry object for this form.
    // Geometries may be shared by multiple
    // particle forms.


  protected void createAppearance()
  {
    appearance = new Appearance();
    appearance.setCapability(Appearance.ALLOW_TEXTURE_WRITE);

    ra = new RenderingAttributes();
    ra.setCapability(RenderingAttributes.ALLOW_VISIBLE_WRITE);

    TextureAttributes xa = new TextureAttributes();
    xa.setTextureMode(TextureAttributes.MODULATE);  // REPLACE, MODULATE

    material = new Material();
    material.setLightingEnable(false);
    material.setAmbientColor(BGlobal.WHITE_3F);
    material.setDiffuseColor(BGlobal.WHITE_3F);
    material.setEmissiveColor(BGlobal.WHITE_3F);
    material.setSpecularColor(BGlobal.WHITE_3F);

    coloring = new ColoringAttributes();
    coloring.setCapability(ColoringAttributes.ALLOW_COLOR_WRITE);
    coloring.setColor(BGlobal.WHITE_3F);

    appearance.setRenderingAttributes(ra);
    appearance.setTextureAttributes(xa);
    appearance.setMaterial(material);
    appearance.setColoringAttributes(coloring);
  }



  public void setAdditive(boolean isAdditive)
  {
    this.isAdditive = isAdditive;

    int destination = isAdditive ?
      TransparencyAttributes.BLEND_ONE :
      TransparencyAttributes.BLEND_ONE_MINUS_SRC_ALPHA;

    ta = new TransparencyAttributes
      (TransparencyAttributes.BLENDED,
       0f,
       TransparencyAttributes.BLEND_SRC_ALPHA,
       destination);

    ta.setCapability(TransparencyAttributes.ALLOW_VALUE_WRITE);
    appearance.setTransparencyAttributes(ta);
  }

  public boolean isAdditive()
    { return isAdditive; }



  protected static Vector3f translation = new Vector3f();
    // Working variable for the form update methods.
    // Assumes only one thread will be accessing
    // the methods below at a time.

  public void start(BParticle p)
  {
    translation.set(p.p);
    transform.set(translation);
    tg.setTransform(transform);

    if (!isAttached)
    {
    	parent.addChild(root);
      isAttached = true;
    }
    else
      // Already attached; make visible instead.
      ra.setVisible(true);
  }

  public void update(BParticle p, float t, float dt)
  {
    translation.set(p.p);
    transform.setTranslation(translation);
    tg.setTransform(transform);
  }


  public void stop(BParticle p)
  {
    // Hide particle from view.
    ra.setVisible(false);
  }



  public void setColor(Color4f c)
    { coloring.setColor(c.x, c.y, c.z); }

  public void setTexture(Texture t)
    { appearance.setTexture(t); }


  public void setSize(float s)
  {
    transform.setScale(s);
    tg.setTransform(transform);
  }



}