package roboforge.ui.components;

import roboforge.BDebug;
import roboforge.BGlobal;
import roboforge.ui.BApplication;

import roboforge.util.io.*;
import roboforge.util.graphics.BImageCache;

import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.*;
import java.io.*;
import java.awt.*;
import java.util.Vector;



public class BImageTable extends JTable
{

  // Class variables...
  // -------------------------------------------------------------

  protected static final int DEFAULT_ICONSIZE = 50;

  protected static final BBasicFilter FILTER =
    new BBasicFilter(".gif");


  // Member variables...
  // -------------------------------------------------------------

  protected ImageTableModel model;
  
  protected IconCellRenderer renderer = 
    new IconCellRenderer(getBackground(), Color.gray);

  

  // Constructors...
  // -------------------------------------------------------------

  public BImageTable(BUIResources res, String key)
  {
    super();

    model = new ImageTableModel();
    initializeUI(res, key);
    
    setModel(model);
  }

  public void initializeUI(BUIResources res, String key)
  {
    res.initComponent(this, key);

    Integer c = (Integer) res.get(key+ ".columns");
    if (c != null)
      model.setColumnCount(c.intValue());
    else
      model.setColumnCount(1);

    setColumnSelectionAllowed(true);
    setRowSelectionAllowed(true);
    setEnabled(true);

    setTableHeader(null);
    setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    
    setAutoscrolls(true);
  }


  // Accessor methods...
  // -------------------------------------------------------------

  public void setSelectedFile(String file)
  {
    if (model == null)
      return;
    
    int r = model.getRowCount();
    int c = model.getColumnCount();
   
    for (int i=0; i<r; i++)
      for (int j=0; j<c; j++)
      {
        Object o = 
          model.getValueAt(i, j);
        
        if (file.equals(o))
          { selectCell(i, j); return; }
      }
  }
  
  protected void selectCell(int r, int c)
  {
    setRowSelectionInterval(r, r);
    setColumnSelectionInterval(c, c);
    
    Rectangle cell = getCellRect(r, c, false);
    scrollRectToVisible(cell);
    
    repaint();
  }

  public String getSelectedItem()
  {
    int r = getSelectedRow();
    int c = getSelectedColumn();

    if ((r != -1) && (c != -1))
      return (String) getModel().getValueAt(r, c);
    else
      return null;
  }
  

  public void setSourcePath(String path)
    throws FileNotFoundException
  {
    model.setPath(path);
    
    // Setup the correct sizes for table cells..
    configureTableSizes();
  }
  
  protected void configureTableSizes()
  {
    // Calculate size of icons...
    int iconSize = getWidth() / model.getColumnCount();

    if (iconSize == 0)
      iconSize = DEFAULT_ICONSIZE;

    configureIcons(iconSize);
    configureCellHeight(iconSize);
    configureCellWidth(iconSize);
  }

  protected void configureCellHeight(int size)
    { setRowHeight(Math.max(size, 1)); }

  protected void configureCellWidth(int size)
  {
    // Set column widths...
    TableColumnModel cModel = getColumnModel();

    if (cModel == null)
      return;

    int cols = cModel.getColumnCount();

    // Configure column widths...
    int available = getWidth();
    TableColumn t;

    for (int i=0; i<(cols-1); i++)
    {
      t = cModel.getColumn(i);
      t.setWidth(size);
      available -= size;
    }

    if (cols > 0)
    {
      t = cModel.getColumn(cols-1);
      t.setWidth(available-2);
    }
  }

  protected void configureIcons(int size)
    { model.reshapeIcons(size); }


  public void reshape(int x, int y, int w, int h)
  {
    super.reshape(x, y, w, h);

    if (model != null)
      configureTableSizes();
  }


  public void addListSelectionListener(ListSelectionListener l)
  {
    ListSelectionModel csm = getColumnModel().getSelectionModel();
    csm.addListSelectionListener(l);

    ListSelectionModel rsm = getSelectionModel();
    rsm.addListSelectionListener(l);
  }
  
  public TableCellRenderer getCellRenderer(int row, int col)
    { return renderer; }
  
  public void setHighlightColor(Color c)
  { 
    renderer.setHighlightColor(c); 
    repaint();
  }


  // -------------------------------------------------------------
  // Image table model Inner class...
  // -------------------------------------------------------------
  
  protected static class ImageTableModel extends AbstractTableModel
  {
    protected static final int MINIMUM_ROWS = 4;

    protected int rows = 0;
    protected int cols = 1;

    protected String[][]            files;
    private   BTranslucentIcon[][]  icons;
    private   BImageCache           cache;

    
    public ImageTableModel()
    {
      cache = BImageCache.getSharedInstance();
    }

    public void setPath(String path) throws FileNotFoundException
    {
      BApplication.setBusy(true);
      
      File file = new File(path);

      if (!file.canRead())
        throw new FileNotFoundException(
          "ImageTableModel.setPath: path ["+path+"] is not valid!");

      String[] titles = file.list(FILTER);
      rows = (titles.length / cols);

      if (rows < cols)
        rows = MINIMUM_ROWS;
      else if ((titles.length % cols) != 0)
        rows++;

      files = new String[rows][cols];
      icons = new BTranslucentIcon[rows][cols];

      int index = 0;
      int n = titles.length;

      for (int i=0; i<rows; i++)
        for (int j=0; j<cols; j++)
          if (index < n)
          {
            files[i][j] = titles[index++];
            icons[i][j] = getFileIcon(path + "/" + files[i][j]);
          }

      // Notify interested parties...
      fireTableStructureChanged();
      fireTableDataChanged();
      
      BApplication.setBusy(false);
    }

    protected BTranslucentIcon getFileIcon(String name)
    {
      Image image = cache.lookup(name);
      
      BTranslucentIcon result =        
        new BTranslucentIcon(image, 
          DEFAULT_ICONSIZE, DEFAULT_ICONSIZE);

      return result;
    }

    
    
    public int getRowCount()
      { return rows; }

    public int getColumnCount()
      { return cols; }
    
    public void setColumnCount(int cols)
      { this.cols = Math.max(1, cols); }

    public Object getValueAt(int row, int column)
    {
      if (files != null)
        return files[row][column];
      else
        return null;
    }
    
    public BTranslucentIcon getIconAt(int row, int column)
    { 
      if (icons != null)
        return icons[row][column]; 
      else
        return null;
    }
    
    public boolean isEmpty()
      { return (rows == 0); }

    
    public void reshapeIcons(int size)
    {
      for (int i=0; i<rows; i++)
        for (int j=0; j<cols; j++)
          if (icons[i][j] != null)
          {
            icons[i][j].setIconWidth(size);
            icons[i][j].setIconHeight(size);
          }
    }

    
  }


  // -------------------------------------------------------------
  // Icon Cell Renderer inner class...
  // -------------------------------------------------------------
  
  protected class IconCellRenderer extends JLabel
    implements TableCellRenderer
  {

    protected Color backgroundColor = Color.black;
    protected Color highlightColor  = Color.white;

    public IconCellRenderer(Color backgroundColor, Color highlightColor)
    {
      this.backgroundColor = backgroundColor;
      this.highlightColor  = highlightColor;
      setOpaque(true);
    }
    
    public void setHighlightColor(Color c)
      { this.highlightColor = c; }

    public Component getTableCellRendererComponent
    (
      JTable table,
      Object value,
      boolean isSelected,
      boolean hasFocus,
      int row,
      int column
    )
    {
      BTranslucentIcon icon = 
        model.getIconAt(row, column);
      
      setIcon(icon);

      if (icon != null)
      {
        if (isSelected)
          icon.setOpacity(0.5f);
        else
          icon.setOpacity(1.0f);
      }

      if (isSelected)
        setBackground(highlightColor);
      else
        setBackground(backgroundColor);

      return this;
    }
  }
  

}