package roboforge.ui.console;

import roboforge.BGlobal;
import roboforge.geometry.BRectangle3D;
import roboforge.BDebug;

import javax.swing.BoundedRangeModel;
import javax.swing.DefaultBoundedRangeModel;
import javax.swing.SwingConstants;
import javax.swing.event.*;

import javax.media.j3d.*;
import javax.vecmath.*;



public class BConsoleProgressBar
  extends BConsoleGeometry
  implements ChangeListener
{

  protected BRectangle3D           backdrop;
  protected BRectangle3D           bar;

  protected Appearance             appearance;
  protected Appearance             backdropAppearance;
  protected Material               material;
  protected Material               backdropMaterial;

  protected ColoringAttributes     foreground;
  protected ColoringAttributes     background;


  protected TransparencyAttributes transparencyAttributes;

  protected Vector3d               progressScale = new Vector3d(1,1,1);
  protected Vector3d               progressTranslate = new Vector3d(0,0,0);
  protected Transform3D            progressScaleTransform = new Transform3D();
  protected TransformGroup         progressScaleGroup = new TransformGroup();

  protected Vector3d               backdropScale = new Vector3d(1,1,1);
  protected Transform3D            backdropScaleTransform = new Transform3D();
  protected TransformGroup         backdropScaleGroup = new TransformGroup();


  protected BoundedRangeModel      model;
  protected int                    alignment;
  protected int                    barWidth;
  protected int                    barHeight;

  protected static final int       X_OFFSET = 0;
  protected static final int       Y_OFFSET = 0;

  protected static final double    MIN_SCALE = 0.001;



  public BConsoleProgressBar(Color3f fg, Color3f bg, int alignment)
    { this(fg, bg, new DefaultBoundedRangeModel(), alignment); }

  public BConsoleProgressBar
    ( Color3f fore, Color3f back, BoundedRangeModel model, int alignment)
  {
    super();

    setModel(model);
    this.alignment = alignment;

    progressScaleGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);

    appearance = new Appearance();

    material = createMaterial();
    appearance.setMaterial(material);

    foreground = createColoring(fore);
    appearance.setColoringAttributes(foreground);

    backdropAppearance = new Appearance();

    TransparencyAttributes backdropTA =
      new TransparencyAttributes(
        TransparencyAttributes.BLENDED ,
        0.7f,
        TransparencyAttributes.BLEND_SRC_ALPHA,
        TransparencyAttributes.BLEND_ONE_MINUS_SRC_ALPHA);

    backdropAppearance.setTransparencyAttributes(backdropTA);

    RenderingAttributes renderAttr = new RenderingAttributes();
    renderAttr.setDepthBufferWriteEnable( false );
    backdropAppearance.setRenderingAttributes(renderAttr);

    backdropMaterial = createMaterial();
    backdropAppearance.setMaterial(backdropMaterial);

    background = createColoring(back);
    backdropAppearance.setColoringAttributes(background);

    backdrop = new BRectangle3D(1, 1);
    backdrop.setAppearance(backdropAppearance);
    backdropScaleGroup.addChild(backdrop);

    bar = new BRectangle3D(1, 1);
    bar.setAppearance(appearance);
    progressScaleGroup.addChild(bar);

    BranchGroup bg = new BranchGroup();
    bg.addChild(progressScaleGroup);
    bg.addChild(backdropScaleGroup);

    setGeometry(bg);
  }

  protected Material createMaterial()
  {
    Material m = new Material();
    m.setDiffuseColor(BGlobal.BLACK_3F);
    m.setAmbientColor(BGlobal.BLACK_3F);
    m.setSpecularColor(BGlobal.BLACK_3F);
    m.setEmissiveColor(BGlobal.WHITE_3F);
    m.setCapability(Material.ALLOW_COMPONENT_WRITE);
    m.setLightingEnable(false);

    return m;
  }

  protected ColoringAttributes createColoring(Color3f c)
  {
    ColoringAttributes ca =
      new ColoringAttributes(c, ColoringAttributes.SHADE_FLAT);

    ca.setCapability(ColoringAttributes.ALLOW_COLOR_WRITE);
    return ca;
  }


  public void setSize(int w, int h)
  {
    super.setSize(w, h);

    setOffset(-w/2, -h/2);

    barWidth  = w - (X_OFFSET*2);
    barHeight = h - (Y_OFFSET*2);

    updateBackdrop();
    updateBarFromModel();
  }


  public void setColor(Color3f c)
  {
    material.setEmissiveColor(c);
    foreground.setColor(c);
  }

  public void setBackgroundColor(Color3f c)
  {
    backdropMaterial.setEmissiveColor(c);
    background.setColor(c);
  }


  public void cleanse()
  {
    super.cleanse();
    appearance = null;
    material = null;
    backdropAppearance = null;
    transparencyAttributes = null;
    backdrop.cleanse();
    bar.cleanse();
  }

  public void stateChanged(ChangeEvent e)
    { updateBarFromModel(); }

  protected void updateBarFromModel()
  {
    double val = (double) model.getValue();
    double min = (double) model.getMinimum();
    double max = (double) model.getMaximum();

    updateBar((val - min) / (max - min));
  }

  private void updateBar(double proportion)
  {
    // Move bar slightly in front of backdrop...
    progressTranslate.z = 0.1;

    // Position bar vertically...
    progressTranslate.y = (this.height / 2) - (barHeight / 2) - Y_OFFSET;

    // Update visual appearance...
    progressScale.x = this.width * proportion;
    progressScale.y = this.height - Y_OFFSET * 2;

    // Ensure no scaling by zero occurs.
    if (progressScale.x < MIN_SCALE)
      progressScale.x = MIN_SCALE;

    progressScaleTransform.setScale(progressScale);

    if (alignment == SwingConstants.LEFT)
      progressTranslate.x = -(1.0 - proportion) / 2.0 * barWidth;

    else if (alignment == SwingConstants.RIGHT)
      progressTranslate.x = (1.0 - proportion) / 2.0 * barWidth;

    progressScaleTransform.setTranslation(progressTranslate);
    progressScaleGroup.setTransform(progressScaleTransform);
  }

  protected void updateBackdrop()
  {
    backdropScale.set(width, height, 1);
    backdropScaleTransform.setScale(backdropScale);
    backdropScaleGroup.setTransform(backdropScaleTransform);
  }


  public void setModel(BoundedRangeModel m)
  {
    if (model != null)
      model.removeChangeListener(this);

    this.model = m;
    model.addChangeListener(this);
  }

  public BoundedRangeModel getModel()
    { return model; }

  public void setValue(int n)
  {
    if (n <getMinimum())
      n = getMinimum();
    else if (n > getMaximum() )
      n =  getMaximum();
      
    model.setValue(n);

  }

  public int getMinimum()
    { return model.getMinimum(); }

  public int getMaximum()
    { return model.getMaximum(); }

}
