package roboforge.ui.modes;

import java.awt.event.MouseEvent;
import java.awt.event.KeyEvent;

import javax.media.j3d.*;
import javax.vecmath.*;
import java.lang.Math.*;

import roboforge.*;
import roboforge.ui.*;
import roboforge.util.timing.*;

////////////////////////////////////////////////////////////////////////////////
//
// BModeArenaAutoCam is a camera mode that follows the bots around the arena
// keeping them both in view
//
//
public class BModeArenaAutoCam extends BModeArena
{
  private BArenaCanvas canvas;

  // the two bots
  private BBot bot1;
  private BBot bot2;

  private double arenaSize;

  private static final double inset = 0.05;

  private static final double fov      = BPhysics.degToRad(75); // 0.5;  // Human Field Of View
  private double oldFov;

  Point3d camPoint = new Point3d();
  Point3d midPoint = new Point3d();
  Point3d botPoint = new Point3d();
  Point3d botPoint2 = new Point3d();

  Vector3d v1 = new Vector3d();
  Vector3d v2 = new Vector3d();
  Vector3d v3 = new Vector3d();

  double flatDis;
  double centerDis;

  double camAngle = 0;
  double angleOffset = 0;
  double heightOffset = 45;

  double radius;

  public BModeArenaAutoCam(BArenaCanvas canvas, double arenaSize)
  {
    super();
    this.canvas = canvas;
    this.arenaSize = arenaSize;
    this.arenaSize = this.arenaSize - (this.arenaSize * inset );
    this.radius = BGlobal.WORLD_TO_CANVAS_SCALE * this.arenaSize;
    this.radius = this.radius - inset;
  }

  public void onActivate()
  {
    super.onActivate();
    oldFov = canvas.getFov();
    canvas.setFov(fov);
    canvas.setConsoleVisible(true);

    updateCamera();
  }

  public void onDeactivate()
  {
    super.onDeactivate();
    canvas.setFov(oldFov);
    canvas.setConsoleVisible(true);
  }


  // do a slow automatic arena animation
  public void animationStep(double t, double dt)
  {
  	double delta  = Math.min(dt * 100, 2.0);
    angleOffset  += delta / 5.0;

    if (angleOffset >= 360.0)
    	angleOffset = angleOffset - 360.0;

    updateCamera(bot1, bot2);
  }


  public void setBots(  BBot bot1, BBot bot2  )
  {
    this.bot1 = bot1;
    this.bot2 = bot2;
  }

  public final void updateCamera()
    { updateCamera (bot1, bot2) ;}


  private synchronized final void updateCamera(BBot bot1, BBot bot2)
  {
    if (bot1 != null && bot2 != null)
    {
      if (bot1.getLiveNode(true) !=  null && bot2.getLiveNode(true) !=  null)
      {
        // set camera view automatically...
        // YLY not needed...
        //botPoint.set(0,0,0);


        BPhysics.getMidPoint(midPoint, bot1, bot2);
        BPhysics.getMidPoint(botPoint, bot1, bot1);
        BPhysics.getMidPoint(botPoint2, bot2, bot2);

        double xDis = midPoint.x - botPoint.x;
        double zDis = midPoint.z - botPoint.z;

        v1.set(xDis, 0.0, zDis);
        v2.set(xDis, 100.0, zDis);

        double absxDis =  Math.abs(xDis);
        double abszDis =  Math.abs(zDis);

        flatDis = Math.sqrt(absxDis * absxDis + abszDis * abszDis);

        // Choose a direction to view the bots from...
        if (flatDis < 0.0001)
          v3.set(0, 0, 1);
        else
        {
          v3.cross(v1,v2);
          v3.normalize();
        }

        double scaleFact =  2.0 * flatDis;

        // make sure the camera doesn't get too close
        if (scaleFact < 250.0)
          scaleFact = 250.0;

        v3.scale(scaleFact);
        camPoint.set(0,0,0);
        camPoint.add(midPoint, v3);

        camAngle = Math.toDegrees(Math.asin((camPoint.x - midPoint.x) / v3.length()));

        double newDis = v3.length();

        float yVal = (float) Math.sin(Math.toRadians(heightOffset)) * (float)newDis;
        newDis  = (float) Math.cos(Math.toRadians(heightOffset)) * (float)newDis;

        // calculate the camera co-ordinates
        if(botPoint.x > botPoint2.x)
        {
          camAngle += angleOffset;
          camPoint.x = midPoint.x + (newDis * Math.sin(Math.toRadians(camAngle)));
          camPoint.z = midPoint.z - (newDis * Math.cos(Math.toRadians(camAngle)));
        }
        else
        {
          camAngle -= angleOffset;
          camPoint.x = midPoint.x + (newDis * Math.sin(Math.toRadians(camAngle)));
          camPoint.z = midPoint.z + (newDis * Math.cos(Math.toRadians(camAngle)));
        }

        // calculate the distance from the camera to the arena center
        centerDis = Math.sqrt( Math.pow( Math.abs(camPoint.x), 2 ) + Math.pow( Math.abs(camPoint.z), 2 ) );

        double difference = 0;
        double adjustDis = 0;

        // check if the camera is outside the arena - if it is then move it in
        if(centerDis > radius)
        {
          difference = centerDis - radius;
          adjustDis = newDis - difference;

          if(botPoint.x > botPoint2.x)
          {
            camPoint.x = midPoint.x + (adjustDis * Math.sin(Math.toRadians(camAngle)));
            camPoint.z = midPoint.z - (adjustDis * Math.cos(Math.toRadians(camAngle)));
          }
          else
          {
            camPoint.x = midPoint.x + (adjustDis * Math.sin(Math.toRadians(camAngle)));
            camPoint.z = midPoint.z + (adjustDis * Math.cos(Math.toRadians(camAngle)));
          }
          yVal += difference;
        }

        // make sure the camera can't go too high
        if(yVal > 600.0f)
          yVal = 600.0f;

        midPoint.y = midPoint.y + 50;

        // set the camera position and look at
        canvas.setCameraPos((float)camPoint.x, yVal, (float)camPoint.z, true, midPoint);

        //midPoint.y = midPoint.y + 50;
        //canvas.lookAt(midPoint, true);
      }
    }
  }


  protected void incrementAngle(double dx)
  {
    angleOffset += dx;

    if (angleOffset >= 360.0)
      angleOffset = angleOffset - 360.0;
    if (angleOffset < 0.0)
      angleOffset = angleOffset + 360.0;
  }

  protected void incrementHeight(double dy)
  {
    heightOffset += dy;

    if (heightOffset < 1.0)
      heightOffset = 1.0;

    if (heightOffset > 89.0)
      heightOffset = 89.0;
  }


  // enable the user to pan around in the view
  public void mouseDragged(MouseEvent e)
  {
    super.mouseDragged(e);

    incrementAngle((double) dx / 3.0);
    incrementHeight((double) dy / 3.0);
    updateCamera(bot1, bot2);
  }




  protected static final double KEY_DX = 5;
  protected static final double KEY_DY = 1;


  public void keyUpPressed(KeyEvent e)
    { incrementHeight(+KEY_DY); updateCamera(bot1, bot2); }

  public void keyDownPressed(KeyEvent e)
    { incrementHeight(-KEY_DY); updateCamera(bot1, bot2); }

  public void keyLeftPressed(KeyEvent e)
    { incrementAngle(KEY_DX); updateCamera(bot1, bot2); }

  public void keyRightPressed(KeyEvent e)
    { incrementAngle(-KEY_DX); updateCamera(bot1, bot2); }


}


