package roboforge.util.io;

import java.io.*;


public class BBlockInputStream extends FilterInputStream
{
  
  // Member variables...
  // -------------------------------------------------------------

  protected BBlockTransform transform;
  
  protected byte[] processed;
    // Processed buffer.
  
  protected int count;
  protected int size;
  
  protected byte[] intStore = new byte[4];
  
  
  
  // Constructor...
  // -------------------------------------------------------------

  public BBlockInputStream(InputStream in, BBlockTransform t)
  {
    super(in);
    transform = t;
    
    int blocksize = 
      transform.getBlockSize();
    
    if (blocksize <= 0)
      throw new IllegalArgumentException("Blocksize is <= 0.");
      
    processed = new byte[blocksize];  
  
    size  = blocksize;
    count = size;
  }
  

  // Internal methods...
  // -------------------------------------------------------------

  protected boolean isEmpty()
    { return count == size; }
  
  
  protected int readInt() throws IOException
  {
    in.read(intStore);
    
    return
      ((intStore[0] & 0xFF) << 24) +
      ((intStore[1] & 0xFF) << 16) +
      ((intStore[2] & 0xFF) <<  8) +
      ((intStore[3] & 0xFF) <<  0);
  }
  
  protected void refill() throws IOException
  {
    int blocksize = readInt();
    
    byte[] pending = new byte[blocksize];
    in.read(pending);

    byte[] transformed
      = transform.transform(pending, 0, blocksize);

    int len = transformed.length;
    count   = size - len;
    
    System.arraycopy(transformed, 0, processed, count, len);
  }
  
  
  protected int handle(byte[] data, int off, int len) throws IOException
  {
    if (count == size)
      refill();

    int avail  = size - count;
    int amount = Math.min(len, avail);
    
    if (amount > 0)
    {
      System.arraycopy(processed, count, data, off, amount);
      count += amount;
    }

    return amount;
  }
  

  // Public methods...
  // -------------------------------------------------------------
  
  public synchronized int read() throws IOException 
  {
    if (isEmpty())
      refill();
    
    return processed[count++] & 0xff;  
	}
  
  public synchronized int read(byte b[], int off, int len) throws IOException
  {
    if (len == 0)
      return 0;
    
    int handled = 0;
    int amount  = len;
    
    while (len > 0)
    {
      handled = handle(b, off, len);
      
      off += handled;
      len -= handled;
    }
    
    return amount;
  }

  public synchronized int available() throws IOException 
    { return size - count; }

  public boolean markSupported() 
    { return false; }
  
}

